/*
 * Programming in Java, Jacobs University, 2015
 *
 * This file demonstrates how lambda expressions and functional
 * interfaces can be used to process data in a functional way. This
 * file also provides a basic idea what streams are in Java 8.
 *
 * $ java B a c b aa bb a
 * a c b a
 * aa bb
 * a a
 * a a b c
 * aa bb
 * average = 1.3333333333333333
 */

import java.util.function.Predicate;
import java.util.Collection;
import java.util.Arrays;
import java.util.List;
import java.util.ArrayList;
import java.util.stream.Collectors;

public class B
{

    /* 
     * Applying the predicate to a list of objects using a
     * conventional loop.
     */
    public static void printMatchingConventional(List<String> l,
						 Predicate<String> p)
    {
	for (String elem: l) {
	    if (p.test(elem)) {
		System.out.print(elem + " ");
	    }
	}
	System.out.println();
    }
    
    /*
     * Applying the predicate to a list using a stream.
     */
    public static void printMatchingStream(List<String> l,
					   Predicate<String> p)
    {
	l.stream().filter(p).forEach((x) -> System.out.print(x + " "));
	System.out.println();
    }

    /*
     * Applying the predicate to a list using a stream and using
     * a method reference.
     */
    public static void printMatchingStreamMethodReference(List<String> l,
							  Predicate<String> p)
    {
	l.stream().filter(p).map(s -> s + " ").forEach(System.out::print);
	System.out.println();
    }

    public static void main(String[] args)
    {
	List<String> list = Arrays.asList(args);
	
	/*
	 * A generic predicate that can be used to filter objects of a
	 * certain type 
	 */
	Predicate<String> t = s -> true;
	Predicate<String> p = s -> s.length() == 1;
	Predicate<String> q = p.negate();
	Predicate<String> a = s -> s.startsWith("a");
	Predicate<String> x = p.and(a);

	printMatchingConventional(list, p);
	printMatchingStream(list, q);
	printMatchingStreamMethodReference(list, x);

	/*
	 * Applying the predicate to a list to create a new list.
	 * Finally, print the new list in a functional style.
	 */
	List<String> l = list.stream()
	    .filter(p)
	    .sorted()
	    .collect(Collectors.toList());
	l.stream().map(s -> s + " ").forEach(System.out::print);
	System.out.println();

	/*
	 * There is a simpler way to remove elements but then this may
	 * be less efficient since I first create a copy.
	 */

	List<String> g = new ArrayList<String>(list);
	g.removeIf(p);
	g.stream().map(s -> s + " ").forEach(System.out::print);
	System.out.println();

	/*
	 * Finally, lets do some mapping that transforms objects of a
	 * stream into different objects and a reduction that absorbs
	 * objects of a stream and turns them into a final (optional)
	 * object.
	 */

	list.stream()
	    .mapToInt(s -> s.length())
	    .average()
	    .ifPresent((d) -> System.out.println("average = " + d));
    }
}
